use std::collections::HashMap;
use flate2::Compression;
use flate2::write::GzEncoder;
use std::path::Path;
extern crate structure;

/**
 * A collection of methods to encode and decode records according to
 * a signature.
 *
 * @author Ishaan Aggarwal
 */

/**
 * If not specified when constructing the instance, than this
 * will be the seekquarry\yioop\library\compressor\Compressor
 * used to compress rows.
 */
// pub const DEFAULT_COMPRESSOR  C\NS_COMPRESSORS . "NonCompressor";
/**
 * If not specified by the format then this will be the assumed fixed
 * length of the primary key
 */
pub const DEFAULT_KEY_LEN: i32 = 16;
/**
 * Array of synonyms for the different possible column name types.
 */
pub const TYPE_SYNONYMS: HashMap<&str, &str> = [ ("BOOLEAN", "BOOL"), ("BOOL", "BOOL"),
    ("CLOB", "TEXT"), ("DOUBLE", "DOUBLE"), ("FLOAT", "REAL"),
("INT", "INT"), ("INTEGER", "INT"), ("REAL", "REAL"),
("TEXT", "TEXT")].iter().cloned().collect();
/**
 * Constant used to indicate that add() should be adding row to an
 * in memory table
 */
pub const ADD_MEM_TABLE: i32 = 0;
/**
 * Constant used to indicate that add() should be adding row using
 * a file handle
 */
pub const ADD_FILE_HANDLE: i32 = 1;
/**
 * Constant used to indicate that add() should be append row row
 * a specified file (not having a file handle to it yet)
 */
pub const ADD_FILE_PATH: i32 = 2;
/**
 * Constant used to indicate that add() should replace the value if
 * there already a row in the table with the same primary key
 */
pub const REPLACE_MODE: i32 = 0;
/**
 * Constant used to indicate that add() should cluster the value if
 * there already a row in the table with the same primary key
 */
const APPEND_MODE: i32 = 1;

enum ValData {
    Type(&str),
    Len(i32)
}

pub struct PackedTableTools {
    /**
     * A string compression algorithm used to compress rows respresented as
     * strings
     * @var seekquarry\yioop\library\compressor\Compressor
     */
    // pub compressor: ,

    /**
     * This is the signature of the records this PackdTableTools will
     * manipulate. This should be an associative array of element of one of
     * the forms: "PRIMARY KEY" => column_name, "PRIMARY KEY" =>
     * [column_name, length_of_primary_key],  column_name => column_type pairs
     * column_type's are from among BOOL, TEXT, DOUBLE, REAL or their
     * synonyms as given in PackedTableTools::TYPE_SYNONYMS. There should
     * be only one primary key pair and if its value is not an array the
     * key length is assumed to be PackedTableTools::DEFAULT_KEY_LEN.
     * @var array
     */
    pub format: HashMap<&str, Vec<ValData>>,

    /**
     * Name of the column used for the primary key.
     * @var string
     */
    pub key_field: String,

    /**
     * Fixed number of bytes used to store the primary key.
     * @var int
     */
    pub key_len: i32,

    /**
     * Number of columns in a record that are of type BOOL
     * @var int
     */
    pub num_bool_columns: i32,

    /**
     * Number of columns in a record that are of type INT
     * @var int
     */
    pub num_int_columns: i32,

    /**
     * Number of columns in a record that are of type TEXT
     * @var int
     */
    pub num_text_columns: i32
}

impl PackedTableTools {

    /**
     * Used to create an instance of a PackedTableTools according to the
     * $format for record columns and $compressor_type to be used for row
     * compression.
     *
     * @param array associative array of items in one of the forms:
     * "PRIMARY KEY" => column_name, "PRIMARY KEY" =>
     * [column_name, length_of_primary_key],  column_name => column_type pairs
     * column_type's are from among BOOL, TEXT, DOUBLE, REAL or their
     * synonyms as given in PackedTableTools::TYPE_SYNONYMS.
     * @param seekquarry\yioop\library\compressor\Compressor used to compress
     *  records
     */
    //TODO: sort out default arguments
    pub fn __construct(format: HashMap<&str, Vec<ValData>>/*, compressor: */ ){
        // compressor = new $compressor_type();
        format = format;
        type_synonyms = self::TYPE_SYNONYMS;
        key_field = null;
        key_len = self::DEFAULT_KEY_LEN;
        num_bool_columns = 0;
        num_int_columns = 0;
        num_text_columns = 0;
        for (key, value) in format {
            if key.to_uppercase().eq("PRIMARY KEY")  {
                if value.len()==2 {
                    key_field = value[0];
                    key_len = value[1];
                    //IMPORTANT: Assuming second value in vec was of type Len as defined in val_data
                } else {
                    key_field = value[0];
                }
                continue;
            }
            if !type_synonyms.contains_key(value[0].to_uppercase()){
                return null;
            }
            value[0] = value[0].to_uppercase();
            format.put(key) = type_synonyms.get(value[0]);
            if value[0].eq("BOOL") {
                num_bool_columns += 1;
            } else if value[0].eq("INT") {
                num_int_columns += 1;
            } else if value[0].eq("TEXT") {
                num_text_columns += 1;
            }
        }
        if key_field.is_empty() {
            return null;
        }
    }
    /**
     * Adds ($key, $table_row) as an entry into $table
     * using the adding and replace methods specified
     *
     * @param mixed &$table either an associative array of key =>value
     *  pairs where the keys are strings of length matching this
     *  PackedTableTools sginature and the values have been packed according to
     *  this signature, or the file name of a file containing a serialized
     *  version of such a table, or a file handle to the end of such a file
     * @param string $key a key string of length given by the signature of this
     *      PackedTableTools
     * @param string $table_row a record packed according tot the signature
     *      of this PackedTableTools
     * @param int $add_method one of self::ADD_MEM_TABLE, self::ADD_FILE_HANDLE,
     *      self::ADD_FILE_PATH indicating which of the three possibilities
     *      for $table we have
     * @param int $mode either self::APPEND_MODE or self::REPLACE_MODE
     *      If the former the table_row data will be appended to any data
     *      currently associated with the key, if the latter it will replace
     *      such data
     * @return bool whether or not adding was successful

    pub fn add(/*&$table,*/ key: &str, table_row: &str,
    add_method: i32, mode: i32){
        let mut encode_key =
            if key_len > 0 {key}
            else { (key.chars().count() as char).to_string() + key};
        match add_method {
            // ADD_FILE_PATH => {
            //     seperator = (fsize($table) > 0) ? "\xFF" : "";
            //     $out = $seperator . encode255(encode_key . $table_row);
            //     $out = $this->compressor->compress($out);
            //     return (file_put_contents($table, $out , FILE_APPEND) > 0);
            // }
            // ADD_FILE_HANDLE => {
            //     $seperator = (ftell($ table) > 0)? "\xFF": "";
            //     $out = $ seperator.encode255($ encode_key. $ table_row);
            //     $out = $ this -> compressor -> compress($ out);
            //     return (fwrite($ table, $out) > 0);
            // }
            // ADD_MEM_TABLE:
            _ => {
                if  mode == REPLACE_MODE || !table.contains_key(key) {
                    table.put(key) = table_row;
                } else if mode == APPEND_MODE {
                    table.put(key) = mergeRowValues(table.get(key), table_row);
                }
            }
        }
        return true;
    }
     */
    /**
     *  Merges two rows of items packed according to this packed table tools
     *  into a single row of items.
     *
     * @param string $row_values1 a row (less key)  packed according according
     *      to this packed tabled tools
     * @param string $row_values2 a second row (less key)  packed according
     *      according to this packed tabled tools
     * @return string a merged row consiting of the items in the first row
     *      followed by those in the second.

    pub fn mergeRowValues(row_values1: &str, row_values2: &str){
        let row_values1_pos = 0;
        let row_values1_count = vByteDecode(row_values1, row_values1_pos);
        let row_values2_pos = 0;
        let row_values2_count = vByteDecode(row_values2, row_values2_pos);
        return vByteEncode(row_values1_count + row_values2_count)+
        &row_values1[row_values1_pos..row_values1.chars().count()-1]+
        &row_values2[row_values2_pos..row_values2.chars().count()-1];
    }
     */
    /**
     * Removes $key and any records associated with it from $table
     *
     * @param array& $table array of key => records pairs where the
     *  key is a string of length given by this PackedTableTool's signature
     *  and the records are packed according this PackedTableTool's signature
     * @param string $key to remove records for
     * @return bool success or failure of removal

    pub fn delete(/*&$table*/ table: HashMap<&str, &str>, key: &str){
        if table.contains_key(key) {
            table.remove(key);
            return true;
        }
        return false;
    }
     */
    /**
     * Return any records in $table associated with $key
     *
     * @param array $table array of key => records pairs where the
     *  key is a string of length given by this PackedTableTool's signature
     *  and the records are packed according this PackedTableTool's signature
     * @param string $key to return records for
     * @return array of records that have been packed according to this
     *  PackedTableTool's signature

    pub fn find(table: HashMap<&str, &str>, key: &str){
        return table.contains_key(key);
    }
     */
    /**
     * Reads in the file $table_path and unpacks it according to this
     * PackedTableTool's signature. Repeated values for a key are handled
     * according to $mode.
     *
     * @param string $table_path path to file containing  PackedTableTool's
     *      records
     * @param int $mode one of self::REPLACE_MODE (replace previous value)
     *      or self::APPEND_MODE (make an array of all values with same key)
     * @return array of key => packed_records pairs

    pub fn load(table_path: &str, mode: i32){
        if !Path::new(table_path).exists() {
            return null;
        }
        let mut table_file = File::open(table_path).expect("Unable to open the file");
        let mut pre_table = String::new();
        table_file.read_to_string(&mut pre_table).expect("Unable to read the file");

        let mut table = compressor.uncompress(pre_table);
        let mut rows: HashMap<&str, &str> = HashMap::new();
        let mut fixed_length_key = key_len > 0;
        let mut replace_mode = mode == REPLACE_MODE;
        let default_key_len = key_len;
        let mut key_offset = if fixed_length_key {0} else {1};
        let more_rows = true;
        let mut cur_start_row = 0;
        while more_rows {
            let mut next_ff_pos = table[0..cur_start_row].find("\xFF".as_str());
            if !next_ff_pos {
                next_ff_pos = table.chars().count();
                more_rows = false;
            }
            let mut entry = decode255(&table[cur_start_row..next_ff_pos]);//TODO: decode
            cur_start_row = next_ff_pos + 1;
            let mut key_len = if fixed_length_key {default_key_len}
                else {entry.chars().next().unwrap() as i32};
            let mut key = entry[key_offset..key_len+key_offset];
            let mut new_values = entry[key_len..key_offset+key_len];
            rows.put(key) = if replace_mode || !rows.contains_key(key) {new_values}
                else {mergeRowValues(rows.get(key), new_values)};
        }
        return rows;
    }
     */
    /**
     * Packs as a string an array of records. Here each records should be an
     * associative array of field items, with field names and types
     * according to this packed table tools signature. The format of the
     * packed records string is: output of bool columns as bit string in
     * order of columns as appear in signature, bit data on sizes to use
     * for each int column (for each column two bit code 00 - 1byte, 01 - 2byte,
     * 10 - 4byte, 11 - 8byte), text column length (1byte/column saying how
     * long the data stored in that column is), This is followed by
     * the actual column data (except bool columns) in the order it is listed
     * in the signature. Int's use their high order bit as a sign bit and are
     * stored using the number of bytes give by their code in the int column
     * bit data. Real/doubles are stored as 8byte doubles.
     *
     * @param array $items array of records to pack
     * @return string records packed into the string format used by
     *      PackedTableTools
     */
    //TODO: need to test items type, pack function
    pub fn pack(&self, items:Vec<HashMap<&str, &str>>){
        format = self.format;
        // if (!isset($items[0])) {
        //     $items = [$items];
        // }//TODO: figure out purpose

        let packed_items = vByteEncode(items.len());
        for item in items.iter() {
            let mut bool_column_data = "";
            let mut cur_bool_int = -1;
            let mut cur_bool_shift = 7;
            let mut int_column_types = "";
            let mut text_column_lengths = "";
            let mut shift = 6;
            let mut cur_int_char = 0;
            let mut int_occurred = false;
            let mut packed_data = "";
            for (key, value) in format {
                if !item.contains_key(key) {
                    return null;
                }
                match value{
                    "BOOL" => {
                        cur_bool_int = if cur_bool_int == -1 {0} else {cur_bool_int};
                        let mut bool_value = if item.get(key) {1} else {0};
                        cur_bool_int += bool_value << cur_bool_shift;
                        cur_bool_shift -= 1;
                        if cur_bool_shift < 0 {
                            bool_column_data = bool_column_data + (cur_bool_int as char).to_string();
                            cur_bool_shift = 7;
                            cur_bool_int = -1;
                        }
                        break;
                    }
                    "DOUBLE" => {
                        //rust equivalent of php pack function
                        let mut rust_pack = structure!(">d");
                        let buf: Vec<u8> = rust_pack.pack(item.get(key));
                        let mut rust_pack_str = str::from_utf8(&buf).unwrap();
                        packed_data = packed_data + rust_pack_str;
                        break;
                    }
                    "INT" => {
                        int_occurred = true;
                        let mut magnitude = item.get(key).abs();
                        let mut positive = if item.get(key) == magnitude {0} else {128};
                        if magnitude < 128 {
                            packed_data = packed_data + ((magnitude + positive) as char).to_string();
                            cur_int_char = if cur_int_char == -1 {0} else {cur_int_char};
                        } else {
                            let mut curr_int_add = 0;
                            let mut packed_int = "";
                            if magnitude < 32768 {
                                let mut rust_pack = structure!(">H");
                                let buf: Vec<u8> = rust_pack.pack(magnitude);
                                packed_int = str::from_utf8(&buf).unwrap();
                                cur_int_add = 1 << shift;
                            } else if magnitude < 2147483647 {
                                let mut rust_pack = structure!(">i");
                                let buf: Vec<u8> = rust_pack.pack(magnitude);
                                packed_int = str::from_utf8(&buf).unwrap();
                                cur_int_add = 2 << shift;
                            } else {
                                let mut rust_pack = structure!(">Q");
                                let buf: Vec<u8> = rust_pack.pack(magnitude);
                                packed_int = str::from_utf8(&buf).unwrap();
                                cur_int_add = 3 << shift;
                            }
                            if !positive {
                                packed_int =
                                    ((packed_int.chars().next().unwrap() as i32) + positive as char).to_string() + packed_int;
                            }
                            packed_data = packed_data + packed_int;
                            cur_int_char = if cur_int_char >= 0 {cur_int_char} else {0};
                            cur_int_char += cur_int_add;
                        }
                        shift -= 2;
                        if shift < 0 {
                            int_column_types = int_column_types + (cur_int_char as char).to_strint();
                            cur_int_char = -1;
                            shift = 6;
                        }
                        break;
                    }
                    "REAL" => {
                        let mut rust_pack = structure!(">f");
                        let buf: Vec<u8> = rust_pack.pack(item.get(key));
                        let mut rust_pack_str = str::from_utf8(&buf).unwrap();
                        packed_data = packed_data + rust_pack_str;
                        break;
                    }
                    "TEXT" => {
                        let mut len = item.get(key).chars().count();
                        if len > 255 {
                            return null;
                        }
                        packed_data = packed_data + item.get(key);
                        text_column_lengths = text_column_lengths +  (len as char).to_string();
                        break;
                    }
                }
            }
            if cur_bool_int != -1 {
                bool_column_data = bool_column_data + (cur_bool_int as char).to_strint();
            }
            if cur_int_char != -1 && int_occurred {
                int_column_types = int_column_types + (cur_int_char as char).to_string();
            }
            packed_items = packed_items + bool_column_data + int_column_types
                + text_column_lengths + packed_data;
        }
        return packed_items;
    }
    /**
     * Saves an associative array of key => packed_records to a file
     * $table_path on disk. Here the packed_records should be packed according
     * to this PackedTableTools signature. Each key value pair is concatenated
     * if variable length keys are used by this packed table tool then the
     * length of the key is stored as a char before the key. Records are then
     * encoded by replacing \xFE => \xFE\xFD and \xFF => \xFE\xFE. The
     * records are then concatenated separated by \xFF, and the result
     * compressed using the PackedTableTool's compressor.
     *
     * @param string $table_path filename to save to
     * @param array $table array of key => packed_records
     * @return bool success or not

    pub fn save($table_path, $table){
        $out = "";
        if ($this->key_len > 0) {
            foreach ($table as $key => $row) {
                $out .= "\xFF" . encode255($key . $row);
            }
        } else {
            foreach ($table as $key => $row) {
                $out .= "\xFF" . encode255(chr(strlen($key)) . $key . $row);
            }
        }
        $out = $this->compressor->compress(substr($out, 1));
        return (file_put_contents($table_path, $out) > 0);
    }
     */
    /**
     * Given a table_row, which might represent several items grouped because
     * share a key, returns the total number of items stored in the row
     * @param string $table_row sequence of items packed according to this
     *      PackedTableTools signature
     * @return int number of items sctored in the table row

    pub fn count(table_row: &str){
        let current_pos = 0;
        return vByteDecode(table_row, current_pos);
    }
     */
    /**
     * Given a table_row, which might represent several items grouped because
     * share a key, unpacks and returns the $offset through $limit numbered
     * items
     * @param string $table_row sequence of items packed according to this
     *      PackedTableTools signature
     * @param int $offset index of item to begin with
     * @param int $limit maximum number of items to return starting at offset
     * @return array unplacked items
     */
    //TODO: sort out default arguments for offset and limit, test unpack function
    pub fn unpack(&self, table_row: &str, offset: i32, limit: i32){
        num_int_columns = self.num_int_columns;
        num_text_columns = self.num_text_columns;
        num_bool_columns = self.num_bool_columns;
        let mut bool_info_len = (self.num_bool_columns/8).ceil() as i32;
        let mut int_info_len = (num_int_columns/4).ceil() as i32;
        let mut unpack_code: HashMap<i32, &str> = [(1, ">H"), (2, ">I"), (3, ">Q")].iter().cloned().collect();
        let mut current_pos = 0;
        let mut num_items = vByteDecode(table_row, current_pos);
        if offset >= num_items {
            return [];
        }
        limit = if limit <= 0 {num_items} else {limit};
        num_items = if limit + offset <= num_items {limit + offset} else {num_items};
        for i in 0..num_items{
            let mut bool_info = table_row[current_pos..current_pos+bool_info_len];
            current_pos += bool_info_len;
            let mut int_info = table_row[current_pos..current_pos+int_info_len];
            current_pos += int_info_len;
            let mut text_info = table_row[current_pos..current_pos+num_text_columns];
            current_pos += num_text_columns;
            let mut format = self.format;
            let mut item: HashMap<&str, &str> = HashMap::new();
            let mut current_bool_pos = 0;
            let mut cur_bool_shift = 7;
            let mut current_int_pos = 0;
            let mut current_text_pos = 0;
            let mut shift = 6;
            let mut bools_used = 0;
            let mut ints_used = 0;
            for (key, value) in format {
                match value {
                    "BOOL" => {
                        if bools_used >= num_bool_columns {
                            return null;
                        }
                        item.put(key) = (((bool_info.chars().skip(current_bool_pos).take(1).collect() as i32) &
                            (1 << cur_bool_shift)) >> cur_bool_shift) > 0;
                        bools_used += 1;
                        cur_bool_shift -= 1;
                        if cur_bool_shift < 0 {
                            current_bool_pos += 1;
                            cur_bool_shift = 7;
                        }
                        break;
                    }
                    "DOUBLE" => {
                        //Rust equivalent of php unpack function
                        let mut rust_unpack = structure!(">d");
                        let buf: Vec<u8> = table_row[current_pos..current_pos+8].as_bytes().to_vec();
                        item.put(key) = rust_unpack.unpack(buf)[0];
                        current_pos += 8;
                        break;
                    }
                    "INT" => {
                        if ints_used >= num_int_columns {
                            return null;
                        }
                        let mut int_code = ((int_info.chars().skip(current_int_pos).take(1).collect() as i32) &
                            (3 << shift)) >> shift;
                        if table_row.chars().count() <= current_pos {
                            return null;
                        }
                        let mut first_char = table_row.chars().skip(current_pos).take(1).collect() as i32;
                        let mut len = 1 << int_code;
                        if int_code == 0 {
                            let mut int_code_value = first_char;
                            if int_code_value > 127 {
                                int_code_value = -(int_code_value & 127);
                            }
                            item.put(key) = int_code_value;
                        } else {
                            let mut sign = if first_char > 127 {-1} else {1};
                            if sign < 0 {
                                table_row.replace_range(current_pos..current_pos+1, ((first_char - 128) as char).to_string);
                            }
                            let mut pre_int = table_row[current_pos..current_pos+len];
                            let mut rust_unpack = structure!(unpack_code.get(int_code));
                            let buf: Vec<u8> = pre_int.as_bytes().to_vec();
                            item.put(key) = sign * rust_unpack(buf)[0];
                        }
                        current_pos += len;
                        ints_used += 1;
                        shift -= 2;
                        if shift < 0 {
                            current_int_pos += 1;
                            shift = 6;
                        }
                        break;
                    }
                    "REAL" => {
                        let mut rust_unpack = structure!(">f");
                        let buf: Vec<u8> = table_row[current_pos..current_pos+4].as_bytes().to_vec();
                        item.put(key) = rust_unpack.unpack(buf)[0];
                        current_pos += 4;
                        break;
                    }
                    "TEXT" => {
                        if current_text_pos >= num_text_columns {
                            return null;
                        }
                        let mut text_len = text_info.chars().skip(current_text_pos).take(1).collect() as i32;
                        item.put(key) = table_row[current_pos..current_pos+text_len];
                        current_pos += text_len;
                        current_text_pos += 1;
                        break;
                    }
                }
            }
            if i >= offset {
            //     items[] = item;
            }//TODO: figure out purpose
        }
        return items;
    }

    /**
     * Decodes from a string using variable byte coding an integer.
     *
     * @param string $str string to use for decoding
     * @param int &$offset byte offset into string when var int stored
     * @return int the decoded integer
     */
    pub fn vByteDecode(str: &str, offset: i32)
    {
        let mut pos_int = (str.chars().skip(offset).take(1).collect() as i32) & 127;
        let mut shift = 7;
        while ((str.chars().skip(offset).take(1).collect() as i32) & 128) > 0 {
            offset += 1;
            pos_int += ((str.chars().skip(offset).take(1).collect() as i32) & 127) << shift;
            shift += 7;
        }
        return pos_int;
    }
}